@@ -6,9 +6,10 @@ from urllib.parse import urlparse
 import dns.resolver
 
 from ..configuration import Configuration
+from ..credentials import OAuth2Credentials
 from ..errors import AutoDiscoverFailed, AutoDiscoverCircularRedirect, TransportError, RedirectError, UnauthorizedError
 from ..protocol import Protocol, FailFast
-from ..transport import get_auth_method_from_response, DEFAULT_HEADERS, NOAUTH
+from ..transport import get_auth_method_from_response, DEFAULT_HEADERS, NOAUTH, OAUTH2, CREDENTIALS_REQUIRED
 from ..util import post_ratelimited, get_domain, get_redirect_url, _back_off_if_needed, _may_retry_on_error, \
     is_valid_hostname, DummyResponse, CONNECTION_ERRORS, TLS_ERRORS
 from ..version import Version
@@ -75,7 +76,7 @@ class Autodiscovery:
         self._emails_visited = []  # Collects Autodiscover email redirects
 
     def discover(self):
-        self._emails_visited.append(self.email)
+        self._emails_visited.append(self.email.lower())
 
         # Check the autodiscover cache to see if we already know the autodiscover service endpoint for this email
         # domain. Use a lock to guard against multiple threads competing to cache information.
@@ -136,7 +137,9 @@ class Autodiscovery:
         # Get the server version. Not all protocol entries have a server version so we cheat a bit and also look at the
         # other ones that point to the same endpoint.
         for protocol in ad_response.account.protocols:
-            if protocol.ews_url.lower() == ews_url.lower() and protocol.server_version:
+            if not protocol.ews_url or not protocol.server_version:
+                continue
+            if protocol.ews_url.lower() == ews_url.lower():
                 version = Version(build=protocol.server_version)
                 break
         else:
@@ -228,6 +231,7 @@ class Autodiscovery:
             with AutodiscoverProtocol.raw_session() as s:
                 try:
                     r = getattr(s, method)(**kwargs)
+                    r.close()  # Release memory
                     break
                 except TLS_ERRORS as e:
                     # Don't retry on TLS errors. They will most likely be persistent.
@@ -268,7 +272,7 @@ class Autodiscovery:
         try:
             session = protocol.get_session()
             r, session = post_ratelimited(protocol=protocol, session=session, url=protocol.service_endpoint,
-                                          headers=headers, data=data, allow_redirects=False)
+                                          headers=headers, data=data, allow_redirects=False, stream=False)
             protocol.release_session(session)
         except UnauthorizedError as e:
             # It's entirely possible for the endpoint to ask for login. We should continue if login fails because this
@@ -286,6 +290,11 @@ class Autodiscovery:
         log.debug('Attempting to get a valid response from %s', url)
         try:
             auth_type, r = self._get_unauthenticated_response(url=url)
+            if isinstance(self.credentials, OAuth2Credentials):
+                # This type of credentials *must* use the OAuth auth type
+                auth_type = OAUTH2
+            elif self.credentials is None and auth_type in CREDENTIALS_REQUIRED:
+                raise ValueError('Auth type %r was detected but no credentials were provided' % auth_type)
             ad_protocol = AutodiscoverProtocol(
                 config=Configuration(
                     service_endpoint=url,
