@@ -122,13 +122,18 @@ class OAuth2Credentials(BaseCredentials):
     :param client_id: ID of an authorized OAuth application
     :param client_secret: Secret associated with the OAuth application
     :param tenant_id: Microsoft tenant ID of the account to access
+    :param identity: An Identity object representing the account that these
+    credentials are connected to.
     """
 
-    def __init__(self, client_id, client_secret, tenant_id):
+    def __init__(self, client_id, client_secret, tenant_id, identity=None):
         super().__init__()
         self.client_id = client_id
         self.client_secret = client_secret
         self.tenant_id = tenant_id
+        self.identity = identity
+        # When set, access_token is a dict (or an oauthlib.oauth2.OAuth2Token, which is also a dict)
+        self.access_token = None
 
     def refresh(self, session):
         # Creating a new session gets a new access token, so there's no
@@ -148,17 +153,28 @@ class OAuth2Credentials(BaseCredentials):
         """
         # Ensure we don't update the object in the middle of a new session
         # being created, which could cause a race
+        if not isinstance(access_token, dict):
+            raise ValueError("'access_token' must be an OAuth2Token")
         with self.lock:
             self.access_token = access_token
 
     def _get_hash_values(self):
-        # access_token is a dict (or an oauthlib.oauth2.OAuth2Token,
-        # which is also a dict) and isn't hashable. Extract its
-        # access_token field, which is the important one.
-        return (
-            getattr(self, k) if k != 'access_token' else self.access_token['access_token']
-            for k in self.__dict__.keys() if k != '_lock'
-        )
+        # 'access_token' may be refreshed once in a while. This should not affect the hash signature.
+        # 'identity' is just informational and should also not affect the hash signature.
+        return (getattr(self, k) for k in self.__dict__.keys() if k not in ('_lock', 'identity', 'access_token'))
+
+    def sig(self):
+        # Like hash(self), but pulls in the access token. Protocol.refresh_credentials() uses this to find out
+        # if the access_token needs to be refreshed.
+        res = []
+        for k in self.__dict__.keys():
+            if k in ('_lock', 'identity'):
+                continue
+            if k == 'access_token':
+                res.append(self.access_token['access_token'] if self.access_token else None)
+                continue
+            res.append(getattr(self, k))
+        return hash(tuple(res))
 
     def __repr__(self):
         return self.__class__.__name__ + repr((self.client_id, '********'))
@@ -200,6 +216,8 @@ class OAuth2AuthorizationCodeCredentials(OAuth2Credentials):
     def __init__(self, client_id=None, client_secret=None, authorization_code=None, access_token=None):
         super().__init__(client_id, client_secret, tenant_id=None)
         self.authorization_code = authorization_code
+        if access_token is not None and not isinstance(access_token, dict):
+            raise ValueError("'access_token' must be an OAuth2Token")
         self.access_token = access_token
 
     def __repr__(self):
