@@ -7,7 +7,7 @@ from decimal import Decimal, InvalidOperation
 import logging
 
 from .errors import ErrorInvalidServerVersion
-from .ewsdatetime import EWSDateTime, EWSDate, EWSTimeZone, NaiveDateTimeNotAllowed, UnknownTimeZone
+from .ewsdatetime import EWSDateTime, EWSDate, EWSTimeZone, NaiveDateTimeNotAllowed, UnknownTimeZone, UTC
 from .util import create_element, get_xml_attrs, set_xml_value, value_to_xml_text, is_iterable, safe_b64decode, TNS
 from .version import Build, Version, EXCHANGE_2013
 
@@ -361,9 +361,10 @@ class FieldURIField(Field):
         return set_xml_value(field_elem, value, version=version)
 
     def field_uri_xml(self):
+        from .properties import FieldURI
         if not self.field_uri:
             raise ValueError("'field_uri' value is missing")
-        return create_element('t:FieldURI', attrs=dict(FieldURI=self.field_uri))
+        return FieldURI(field_uri=self.field_uri).to_xml(version=None)
 
     def request_tag(self):
         if not self.field_uri_postfix:
@@ -575,6 +576,30 @@ class DateField(FieldURIField):
         return self.default
 
 
+class DateTimeBackedDateField(FieldURIField):
+    # A field that acts like a date, but where values are sent to EWS as EWSDateTime.
+    value_cls = datetime.date
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        # Create internal field to handle datetime-only logic
+        self._datetime_field = DateTimeField(*args, **kwargs)
+
+    def date_to_datetime(self, value):
+        return UTC.localize(self._datetime_field.value_cls.combine(value, datetime.time(11, 59)))
+
+    def from_xml(self, elem, account):
+        res = self._datetime_field.from_xml(elem=elem, account=account)
+        if res is None:
+            return res
+        return res.date()
+
+    def to_xml(self, value, version):
+        # Convert date to datetime. EWS changes all values to have a time of 11:59 local time, so let's send that.
+        value = self.date_to_datetime(value)
+        return self._datetime_field.to_xml(value=value, version=version)
+
+
 class TimeField(FieldURIField):
     value_cls = datetime.time
 
@@ -623,6 +648,27 @@ class DateTimeField(FieldURIField):
         return self.default
 
 
+class DateOrDateTimeField(DateTimeField):
+    """This field can handle both EWSDate and EWSDateTime. Used for calendar items where 'start' and 'end'
+    values are conceptually dates when the calendar item is an all-day event, but datetimes in all other cases.
+
+    For all-day items, we assume both start and end dates are inclusive.
+
+    For filtering kwarg validation and other places where we must decide on a specific class, we settle on datetime.
+    """
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        # Create internal field to handle date-only logic
+        self._date_field = DateField(*args, **kwargs)
+
+    def clean(self, value, version=None):
+        # Most calendar items will contain datetime values. We can't access the is_all_day value here, so CalendarItem
+        # must handle that sanity check.
+        if type(value) == EWSDate:
+            return self._date_field.clean(value=value, version=version)
+        return super().clean(value=value, version=version)
+
+
 class TimeZoneField(FieldURIField):
     value_cls = EWSTimeZone
 
@@ -643,7 +689,7 @@ class TimeZoneField(FieldURIField):
 
     def to_xml(self, value, version):
         return create_element(
-            't:%s' % self.field_uri_postfix,
+            self.request_tag(),
             attrs=OrderedDict([
                 ('Id', value.ms_id),
                 ('Name', value.ms_name),
@@ -1063,13 +1109,8 @@ class SubField(Field):
 
     @staticmethod
     def field_uri_xml(field_uri, label):
-        return create_element(
-            't:IndexedFieldURI',
-            attrs=OrderedDict([
-                ('FieldURI', field_uri),
-                ('FieldIndex', label),
-            ])
-        )
+        from .properties import IndexedFieldURI
+        return IndexedFieldURI(field_uri=field_uri, field_index=label).to_xml(version=None)
 
     def __hash__(self):
         return hash(self.name)
@@ -1105,13 +1146,8 @@ class NamedSubField(SubField):
         return set_xml_value(field_elem, value, version=version)
 
     def field_uri_xml(self, field_uri, label):
-        return create_element(
-            't:IndexedFieldURI',
-            attrs=OrderedDict([
-                ('FieldURI', '%s:%s' % (field_uri, self.field_uri)),
-                ('FieldIndex', label),
-            ])
-        )
+        from .properties import IndexedFieldURI
+        return IndexedFieldURI(field_uri='%s:%s' % (field_uri, self.field_uri), field_index=label).to_xml(version=None)
 
     def request_tag(self):
         return 't:%s' % self.field_uri
@@ -1150,6 +1186,20 @@ class EmailAddressesField(IndexedField):
     def field_uri_xml(self):
         raise NotImplementedError()
 
+    def clean(self, value, version=None):
+        if value is not None:
+            default_labels = self.value_cls.LABEL_CHOICES
+            if len(value) > len(default_labels):
+                raise ValueError('This field can handle at most %s values (value: %r)' % (len(default_labels), value))
+            tmp = []
+            for s, default_label in zip(value, default_labels):
+                if not isinstance(s, str):
+                    tmp.append(s)
+                    continue
+                tmp.append(self.value_cls(email=s, label=default_label))
+            value = tmp
+        return super().clean(value, version=version)
+
 
 class PhoneNumberField(IndexedField):
     is_list = True
@@ -1198,20 +1248,16 @@ class ExtendedPropertyField(Field):
         return value
 
     def field_uri_xml(self):
-        elem = create_element('t:ExtendedFieldURI')
+        from .properties import ExtendedFieldURI
         cls = self.value_cls
-        if cls.distinguished_property_set_id:
-            elem.set('DistinguishedPropertySetId', cls.distinguished_property_set_id)
-        if cls.property_set_id:
-            elem.set('PropertySetId', cls.property_set_id)
-        if cls.property_tag:
-            elem.set('PropertyTag', cls.property_tag_as_hex())
-        if cls.property_name:
-            elem.set('PropertyName', cls.property_name)
-        if cls.property_id:
-            elem.set('PropertyId', value_to_xml_text(cls.property_id))
-        elem.set('PropertyType', cls.property_type)
-        return elem
+        return ExtendedFieldURI(
+            distinguished_property_set_id=cls.distinguished_property_set_id,
+            property_set_id=cls.property_set_id.lower() if cls.property_set_id else None,
+            property_tag=cls.property_tag_as_hex(),
+            property_name=cls.property_name,
+            property_id=value_to_xml_text(cls.property_id) if cls.property_id else None,
+            property_type=cls.property_type,
+        ).to_xml(version=None)
 
     def from_xml(self, elem, account):
         extended_properties = elem.findall(self.value_cls.response_tag())
@@ -1300,3 +1346,17 @@ class ProtocolListField(EWSElementListField):
 
     def from_xml(self, elem, account):
         return [self.value_cls.from_xml(elem=e, account=account) for e in elem.findall(self.value_cls.response_tag())]
+
+
+class RoutingTypeField(ChoiceField):
+    def __init__(self, *args, **kwargs):
+        kwargs['choices'] = {Choice('SMTP'), Choice('EX')}
+        kwargs['default'] = 'SMTP'
+        super().__init__(*args, **kwargs)
+
+
+class IdElementField(EWSElementField):
+    def __init__(self, *args, **kwargs):
+        kwargs['is_searchable'] = False
+        kwargs['is_read_only'] = True
+        super().__init__(*args, **kwargs)
