@@ -5,14 +5,14 @@ from operator import attrgetter
 from ..errors import ErrorAccessDenied, ErrorFolderNotFound, ErrorCannotEmptyFolder, ErrorCannotDeleteObject, \
     ErrorDeleteDistinguishedFolder
 from ..fields import IntegerField, CharField, FieldPath, EffectiveRightsField, PermissionSetField, EWSElementField, \
-    Field
+    Field, IdElementField
 from ..items import CalendarItem, RegisterMixIn, Persona, ITEM_CLASSES, ITEM_TRAVERSAL_CHOICES, SHAPE_CHOICES, \
     ID_ONLY, DELETE_TYPE_CHOICES, HARD_DELETE, SHALLOW as SHALLOW_ITEMS
-from ..properties import Mailbox, FolderId, ParentFolderId, InvalidField, DistinguishedFolderId
+from ..properties import Mailbox, FolderId, ParentFolderId, InvalidField, DistinguishedFolderId, Fields
 from ..queryset import QuerySet, SearchableMixIn, DoesNotExist
 from ..restriction import Restriction
 from ..services import CreateFolder, UpdateFolder, DeleteFolder, EmptyFolder, FindPeople
-from ..util import TNS
+from ..util import TNS, require_id
 from ..version import Version, EXCHANGE_2007_SP1, EXCHANGE_2010
 from .collections import FolderCollection
 from .queryset import SingleFolderQuerySet, SHALLOW as SHALLOW_FOLDERS, DEEP as DEEP_FOLDERS
@@ -40,7 +40,8 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
     LOCALIZED_NAMES = dict()  # A map of (str)locale: (tuple)localized_folder_names
     ITEM_MODEL_MAP = {cls.response_tag(): cls for cls in ITEM_CLASSES}
     ID_ELEMENT_CLS = FolderId
-    LOCAL_FIELDS = [
+    FIELDS = Fields(
+        IdElementField('_id', field_uri='folder:FolderId', value_cls=ID_ELEMENT_CLS),
         EWSElementField('parent_folder_id', field_uri='folder:ParentFolderId', value_cls=ParentFolderId,
                         is_read_only=True),
         CharField('folder_class', field_uri='folder:FolderClass', is_required_after_save=True),
@@ -48,10 +49,9 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
         IntegerField('total_count', field_uri='folder:TotalCount', is_read_only=True),
         IntegerField('child_folder_count', field_uri='folder:ChildFolderCount', is_read_only=True),
         IntegerField('unread_count', field_uri='folder:UnreadCount', is_read_only=True),
-    ]
-    FIELDS = RegisterMixIn.FIELDS + LOCAL_FIELDS
+    )
 
-    __slots__ = tuple(f.name for f in LOCAL_FIELDS) + ('is_distinguished',)
+    __slots__ = tuple(f.name for f in FIELDS) + ('is_distinguished',)
 
     # Used to register extended properties
     INSERT_AFTER_FIELD = 'child_folder_count'
@@ -127,12 +127,16 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
         elif head == '**':
             # Match anything here or in any subfolder at arbitrary depth
             for c in self.walk():
-                if fnmatch(c.name, tail or '*'):
+                # fnmatch() may be case-sensitive depending on operating system:
+                # force a case-insensitive match since case appears not to
+                # matter for folders in Exchange
+                if fnmatch(c.name.lower(), (tail or '*').lower()):
                     yield c
         else:
             # Regular pattern
             for c in self.children:
-                if not fnmatch(c.name, head):
+                # See note above on fnmatch() case-sensitivity
+                if not fnmatch(c.name.lower(), head.lower()):
                     continue
                 if tail is None:
                     yield c
@@ -362,12 +366,8 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
             # New folder
             if update_fields:
                 raise ValueError("'update_fields' is only valid for updates")
-            res = list(CreateFolder(account=self.account).call(parent_folder=self.parent, folders=[self]))
-            if len(res) != 1:
-                raise ValueError('Expected result length 1, but got %s' % res)
-            if isinstance(res[0], Exception):
-                raise res[0]
-            self.id, self.changekey = res[0].id, res[0].changekey
+            res = CreateFolder(account=self.account).get(parent_folder=self.parent, folders=[self])
+            self._id = self.ID_ELEMENT_CLS(res.id, res.changekey)
             self.root.add_folder(self)  # Add this folder to the cache
             return self
 
@@ -384,12 +384,8 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
                         # These are required and cannot be deleted
                         continue
                 update_fields.append(f.name)
-        res = list(UpdateFolder(account=self.account).call(folders=[(self, update_fields)]))
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        folder_id, changekey = res[0].id, res[0].changekey
+        res = UpdateFolder(account=self.account).get(folders=[(self, update_fields)])
+        folder_id, changekey = res.id, res.changekey
         if self.id != folder_id:
             raise ValueError('ID mismatch')
         # Don't check changekey value. It may not change on no-op updates
@@ -400,24 +396,16 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
     def delete(self, delete_type=HARD_DELETE):
         if delete_type not in DELETE_TYPE_CHOICES:
             raise ValueError("'delete_type' %s must be one of %s" % (delete_type, DELETE_TYPE_CHOICES))
-        res = list(DeleteFolder(account=self.account).call(folders=[self], delete_type=delete_type))
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
+        DeleteFolder(account=self.account).get(folders=[self], delete_type=delete_type)
         self.root.remove_folder(self)  # Remove the updated folder from the cache
-        self.id, self.changekey = None, None
+        self._id = None
 
     def empty(self, delete_type=HARD_DELETE, delete_sub_folders=False):
         if delete_type not in DELETE_TYPE_CHOICES:
             raise ValueError("'delete_type' %s must be one of %s" % (delete_type, DELETE_TYPE_CHOICES))
-        res = list(EmptyFolder(account=self.account).call(
-            folders=[self], delete_type=delete_type, delete_sub_folders=delete_sub_folders)
+        EmptyFolder(account=self.account).get(
+            folders=[self], delete_type=delete_type, delete_sub_folders=delete_sub_folders
         )
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
         if delete_sub_folders:
             # We don't know exactly what was deleted, so invalidate the entire folder cache to be safe
             self.root.clear_cache()
@@ -466,13 +454,9 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
 
     @classmethod
     def _kwargs_from_elem(cls, elem, account):
-        folder_id, changekey = cls.id_from_xml(elem)
-        kwargs = dict(id=folder_id, changekey=changekey)
         # Check for 'DisplayName' element before collecting kwargs because because that clears the elements
         has_name_elem = elem.find(cls.get_field_by_fieldname('name').response_tag()) is not None
-        kwargs.update({
-            f.name: f.from_xml(elem=elem, account=account) for f in cls.FIELDS if f.name not in ('id', 'changekey')
-        })
+        kwargs = {f.name: f.from_xml(elem=elem, account=account) for f in cls.FIELDS}
         if has_name_elem and not kwargs['name']:
             # When we request the 'DisplayName' property, some folders may still be returned with an empty value.
             # Assign a default name to these folders.
@@ -493,6 +477,10 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
             return FolderId(id=self.id, changekey=self.changekey).to_xml(version=version)
         return super().to_xml(version=version)
 
+    def to_id_xml(self, version):
+        # Folder(name='Foo') is a perfectly valid ID to e.g. create a folder
+        return self.to_xml(version=version)
+
     @classmethod
     def resolve(cls, account, folder):
         # Resolve a single folder
@@ -508,11 +496,8 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
             raise ValueError("Expected folder %r to be a %s instance" % (f, cls))
         return f
 
+    @require_id
     def refresh(self):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
         fresh_folder = self.resolve(account=self.account, folder=self)
         if self.id != fresh_folder.id:
             raise ValueError('ID mismatch')
@@ -560,11 +545,11 @@ class BaseFolder(RegisterMixIn, SearchableMixIn):
 
 class Folder(BaseFolder):
     """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/folder"""
-    LOCAL_FIELDS = [
+    LOCAL_FIELDS = Fields(
         PermissionSetField('permission_set', field_uri='folder:PermissionSet', supported_from=EXCHANGE_2007_SP1),
         EffectiveRightsField('effective_rights', field_uri='folder:EffectiveRights', is_read_only=True,
                              supported_from=EXCHANGE_2007_SP1),
-    ]
+    )
     FIELDS = BaseFolder.FIELDS + LOCAL_FIELDS
 
     __slots__ = tuple(f.name for f in LOCAL_FIELDS) + ('_root',)
