@@ -1,13 +1,18 @@
+import datetime
 import logging
 
+from ..ewsdatetime import EWSDate, EWSDateTime
 from ..fields import BooleanField, IntegerField, TextField, ChoiceField, URIField, BodyField, DateTimeField, \
     MessageHeaderField, AttachmentField, RecurrenceField, MailboxField, AttendeesField, Choice, OccurrenceField, \
     OccurrenceListField, TimeZoneField, CharField, EnumAsIntField, FreeBusyStatusField, ReferenceItemIdField, \
-    AssociatedCalendarItemIdField
-from ..properties import Attendee, ReferenceItemId, AssociatedCalendarItemId
+    AssociatedCalendarItemIdField, DateOrDateTimeField
+from ..properties import Attendee, ReferenceItemId, AssociatedCalendarItemId, OccurrenceItemId, RecurringMasterItemId, \
+    Fields
 from ..recurrence import FirstOccurrence, LastOccurrence, Occurrence, DeletedOccurrence
+from ..services import CreateItem
+from ..util import set_xml_value, require_account
 from ..version import EXCHANGE_2010, EXCHANGE_2013
-from .base import BaseItem, BaseReplyItem, SEND_AND_SAVE_COPY
+from .base import BaseItem, BaseReplyItem, BulkCreateResult, SEND_ONLY, SEND_AND_SAVE_COPY, SEND_TO_NONE
 from .item import Item
 from .message import Message
 
@@ -53,10 +58,10 @@ class CalendarItem(Item, AcceptDeclineMixIn):
     MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/calendaritem
     """
     ELEMENT_NAME = 'CalendarItem'
-    LOCAL_FIELDS = [
+    LOCAL_FIELDS = Fields(
         TextField('uid', field_uri='calendar:UID', is_required_after_save=True, is_searchable=False),
-        DateTimeField('start', field_uri='calendar:Start', is_required=True),
-        DateTimeField('end', field_uri='calendar:End', is_required=True),
+        DateOrDateTimeField('start', field_uri='calendar:Start', is_required=True),
+        DateOrDateTimeField('end', field_uri='calendar:End', is_required=True),
         DateTimeField('original_start', field_uri='calendar:OriginalStart', is_read_only=True),
         BooleanField('is_all_day', field_uri='calendar:IsAllDayEvent', is_required=True, default=False),
         FreeBusyStatusField('legacy_free_busy_status', field_uri='calendar:LegacyFreeBusyStatus', is_required=True,
@@ -112,11 +117,35 @@ class CalendarItem(Item, AcceptDeclineMixIn):
                      is_read_only=True),
         URIField('meeting_workspace_url', field_uri='calendar:MeetingWorkspaceUrl'),
         URIField('net_show_url', field_uri='calendar:NetShowUrl'),
-    ]
+    )
     FIELDS = Item.FIELDS + LOCAL_FIELDS
 
     __slots__ = tuple(f.name for f in LOCAL_FIELDS)
 
+    def occurrence(self, index):
+        """Return a new CalendarItem instance with an ID pointing to the n'th occurrence in the recurrence. The index
+        is 1-based. No other field values are fetched from the server.
+
+        Only call this method on a recurring master.
+        """
+        return self.__class__(
+            account=self.account,
+            folder=self.folder,
+            _id=OccurrenceItemId(id=self.id, changekey=self.changekey, instance_index=index),
+        )
+
+    def recurring_master(self):
+        """Return a new CalendarItem instance with an ID pointing to the recurring master item of this occurrence. No
+        other field values are fetched from the server.
+
+        Only call this method on an occurrence of a recurring master.
+        """
+        return self.__class__(
+            account=self.account,
+            folder=self.folder,
+            _id=RecurringMasterItemId(id=self.id, changekey=self.changekey),
+        )
+
     @classmethod
     def timezone_fields(cls):
         return [f for f in cls.FIELDS if isinstance(f, TimeZoneField)]
@@ -124,17 +153,29 @@ class CalendarItem(Item, AcceptDeclineMixIn):
     def clean_timezone_fields(self, version):
         # pylint: disable=access-member-before-definition
         # Sets proper values on the timezone fields if they are not already set
+        if self.start is None:
+            start_tz = None
+        elif type(self.start) == EWSDate:
+            start_tz = self.account.default_timezone
+        else:
+            start_tz = self.start.tzinfo
+        if self.end is None:
+            end_tz = None
+        elif type(self.end) == EWSDate:
+            end_tz = self.account.default_timezone
+        else:
+            end_tz = self.end.tzinfo
         if version.build < EXCHANGE_2010:
-            if self._meeting_timezone is None and self.start is not None:
-                self._meeting_timezone = self.start.tzinfo
+            if self._meeting_timezone is None:
+                self._meeting_timezone = start_tz
             self._start_timezone = None
             self._end_timezone = None
         else:
             self._meeting_timezone = None
-            if self._start_timezone is None and self.start is not None:
-                self._start_timezone = self.start.tzinfo
-            if self._end_timezone is None and self.end is not None:
-                self._end_timezone = self.end.tzinfo
+            if self._start_timezone is None:
+                self._start_timezone = start_tz
+            if self._end_timezone is None:
+                self._end_timezone = end_tz
 
     def clean(self, version=None):
         # pylint: disable=access-member-before-definition
@@ -160,6 +201,62 @@ class CalendarItem(Item, AcceptDeclineMixIn):
             update_fields.remove('uid')
         return update_fields
 
+    @classmethod
+    def from_xml(cls, elem, account):
+        item = super().from_xml(elem=elem, account=account)
+        # EWS returns the start and end values as a datetime regardless of the is_all_day status. Convert to date if
+        # applicable.
+        if not item.is_all_day:
+            return item
+        for field_name in ('start', 'end'):
+            val = getattr(item, field_name)
+            if val is None:
+                continue
+            # Return just the date part of the value. Subtract 1 day from the date if this is the end field. This is
+            # the inverse of what we do in .to_xml(). Convert to the local timezone before getting the date.
+            if field_name == 'end':
+                val -= datetime.timedelta(days=1)
+            tz = getattr(item, '_%s_timezone' % field_name)
+            setattr(item, field_name, val.astimezone(tz).date())
+        return item
+
+    def date_to_datetime(self, field_name):
+        # EWS always expects a datetime. If we have a date value, then convert it to datetime in the local
+        # timezone. Additionally, if this the end field, add 1 day to the date. We could add 12 hours to both
+        # start and end values and let EWS apply its logic, but that seems hacky.
+        value = getattr(self, field_name)
+        if self.account.version.build < EXCHANGE_2010:
+            tz = self._meeting_timezone
+        else:
+            tz = getattr(self, '_%s_timezone' % field_name)
+        value = tz.localize(EWSDateTime.combine(value, datetime.time(0, 0)))
+        if field_name == 'end':
+            value += datetime.timedelta(days=1)
+        return value
+
+    def to_xml(self, version):
+        # EWS has some special logic related to all-day start and end values. Non-midnight start values are pushed to
+        # the previous midnight. Non-midnight end values are pushed to the following midnight. Midnight in this context
+        # refers to midnight in the local timezone. See
+        #
+        # https://docs.microsoft.com/en-us/exchange/client-developer/exchange-web-services/how-to-create-all-day-events-by-using-ews-in-exchange
+        #
+        elem = super().to_xml(version=version)
+        if not self.is_all_day:
+            return elem
+        for field_name in ('start', 'end'):
+            value = getattr(self, field_name)
+            if value is None:
+                continue
+            if type(value) == EWSDate:
+                # EWS always expects a datetime
+                value = self.date_to_datetime(field_name=field_name)
+                # We already generated an XML element for this field, but it contains a plain date at this point, which
+                # is invalid. Replace the value.
+                field = self.get_field_by_fieldname(field_name)
+                set_xml_value(elem=elem.find(field.response_tag()), value=value, version=version)
+        return elem
+
 
 class BaseMeetingItem(Item):
     """
@@ -172,7 +269,7 @@ class BaseMeetingItem(Item):
     Therefore BaseMeetingItem inherits from  EWSElement has no save() or send() method
 
     """
-    LOCAL_FIELDS = Message.LOCAL_FIELDS[:-2] + [
+    LOCAL_FIELDS = Message.LOCAL_FIELDS[:-2] + Fields(
         AssociatedCalendarItemIdField('associated_calendar_item_id', field_uri='meeting:AssociatedCalendarItemId',
                                       value_cls=AssociatedCalendarItemId),
         BooleanField('is_delegated', field_uri='meeting:IsDelegated', is_read_only=True, default=False),
@@ -182,7 +279,7 @@ class BaseMeetingItem(Item):
                     choices={Choice('Unknown'), Choice('Organizer'), Choice('Tentative'),
                              Choice('Accept'), Choice('Decline'), Choice('NoResponseReceived')},
                     is_required=True, default='Unknown'),
-    ]
+    )
     FIELDS = Item.FIELDS + LOCAL_FIELDS
 
     __slots__ = tuple(f.name for f in LOCAL_FIELDS)
@@ -193,7 +290,7 @@ class MeetingRequest(BaseMeetingItem, AcceptDeclineMixIn):
     MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/meetingrequest
     """
     ELEMENT_NAME = 'MeetingRequest'
-    LOCAL_FIELDS = [
+    LOCAL_FIELDS = Fields(
         ChoiceField('meeting_request_type', field_uri='meetingRequest:MeetingRequestType',
                     choices={Choice('FullUpdate'), Choice('InformationalUpdate'), Choice('NewMeetingRequest'),
                              Choice('None'), Choice('Outdated'), Choice('PrincipalWantsCopy'),
@@ -202,7 +299,7 @@ class MeetingRequest(BaseMeetingItem, AcceptDeclineMixIn):
         ChoiceField('intended_free_busy_status', field_uri='meetingRequest:IntendedFreeBusyStatus', choices={
                     Choice('Free'), Choice('Tentative'), Choice('Busy'), Choice('OOF'), Choice('NoData')},
                     is_required=True, default='Busy'),
-    ] + [f for f in CalendarItem.LOCAL_FIELDS[1:] if f.name != 'is_response_requested']
+    ) + Fields(*(f for f in CalendarItem.LOCAL_FIELDS[1:] if f.name != 'is_response_requested'))
 
     # FIELDS on this element are shuffled compared to other elements
     culture_idx = None
@@ -234,10 +331,10 @@ class MeetingMessage(BaseMeetingItem):
 class MeetingResponse(BaseMeetingItem):
     """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/meetingresponse"""
     ELEMENT_NAME = 'MeetingResponse'
-    LOCAL_FIELDS = [
+    LOCAL_FIELDS = Fields(
         MailboxField('received_by', field_uri='message:ReceivedBy', is_read_only=True),
         MailboxField('received_representing', field_uri='message:ReceivedRepresenting', is_read_only=True),
-    ]
+    )
     # FIELDS on this element are shuffled compared to other elements
     culture_idx = None
     for i, field in enumerate(Item.FIELDS):
@@ -261,7 +358,7 @@ class MeetingCancellation(BaseMeetingItem):
 
 class BaseMeetingReplyItem(BaseItem):
     """Base class for meeting request reply items that share the same fields (Accept, TentativelyAccept, Decline)"""
-    FIELDS = [
+    FIELDS = Fields(
         CharField('item_class', field_uri='item:ItemClass', is_read_only=True),
         ChoiceField('sensitivity', field_uri='item:Sensitivity', choices={
             Choice('Normal'), Choice('Personal'), Choice('Private'), Choice('Confidential')
@@ -269,26 +366,26 @@ class BaseMeetingReplyItem(BaseItem):
         BodyField('body', field_uri='item:Body'),  # Accepts and returns Body or HTMLBody instances
         AttachmentField('attachments', field_uri='item:Attachments'),  # ItemAttachment or FileAttachment
         MessageHeaderField('headers', field_uri='item:InternetMessageHeaders', is_read_only=True),
-    ] + Message.LOCAL_FIELDS[:6] + [
+    ) + Message.LOCAL_FIELDS[:6] + Fields(
         ReferenceItemIdField('reference_item_id', field_uri='item:ReferenceItemId', value_cls=ReferenceItemId),
         MailboxField('received_by', field_uri='message:ReceivedBy', is_read_only=True),
         MailboxField('received_representing', field_uri='message:ReceivedRepresenting', is_read_only=True),
         DateTimeField('proposed_start', field_uri='meeting:ProposedStart', supported_from=EXCHANGE_2013),
         DateTimeField('proposed_end', field_uri='meeting:ProposedEnd', supported_from=EXCHANGE_2013),
-    ]
+    )
 
     __slots__ = tuple(f.name for f in FIELDS)
 
+    @require_account
     def send(self, message_disposition=SEND_AND_SAVE_COPY):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-
-        res = self.account.bulk_create(items=[self], folder=self.folder, message_disposition=message_disposition)
-
-        for r_item in res:
-            if isinstance(r_item, Exception):
-                raise r_item
-        return res
+        res = CreateItem(account=self.account).get(
+            items=[self],
+            folder=self.folder,
+            message_disposition=message_disposition,
+            send_meeting_invitations=SEND_TO_NONE,
+            expect_result=message_disposition not in (SEND_ONLY, SEND_AND_SAVE_COPY),
+        )
+        return BulkCreateResult.from_xml(elem=res, account=self)
 
 
 class AcceptItem(BaseMeetingReplyItem):
@@ -315,5 +412,5 @@ class DeclineItem(BaseMeetingReplyItem):
 class CancelCalendarItem(BaseReplyItem):
     """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/cancelcalendaritem"""
     ELEMENT_NAME = 'CancelCalendarItem'
-    FIELDS = [f for f in BaseReplyItem.FIELDS if f.name != 'author']
+    FIELDS = Fields(*(f for f in BaseReplyItem.FIELDS if f.name != 'author'))
     __slots__ = tuple()
