@@ -2,50 +2,17 @@ import logging
 
 from ..fields import BooleanField, IntegerField, TextField, CharListField, ChoiceField, URIField, BodyField, \
     DateTimeField, MessageHeaderField, AttachmentField, Choice, EWSElementField, EffectiveRightsField, CultureField, \
-    CharField, MimeContentField
-from ..properties import ConversationId, ParentFolderId, ReferenceItemId
-from ..util import is_iterable
+    CharField, MimeContentField, FieldPath
+from ..properties import ConversationId, ParentFolderId, ReferenceItemId, OccurrenceItemId, RecurringMasterItemId,\
+    Fields
+from ..services import GetItem, CreateItem, UpdateItem, DeleteItem, MoveItem, CopyItem, ArchiveItem
+from ..util import is_iterable, require_account, require_id
 from ..version import EXCHANGE_2010, EXCHANGE_2013
-from .base import BaseItem, SAVE_ONLY, SEND_ONLY, SEND_AND_SAVE_COPY
+from .base import BaseItem, BulkCreateResult, SAVE_ONLY, SEND_ONLY, SEND_AND_SAVE_COPY, ID_ONLY, SEND_TO_NONE, \
+    AUTO_RESOLVE, SOFT_DELETE, HARD_DELETE, ALL_OCCURRENCIES, MOVE_TO_DELETED_ITEMS
 
 log = logging.getLogger(__name__)
 
-# SendMeetingInvitations values. See
-# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/createitem
-# SendMeetingInvitationsOrCancellations. See
-# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/updateitem
-# SendMeetingCancellations values. See
-# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/deleteitem
-SEND_TO_NONE = 'SendToNone'
-SEND_ONLY_TO_ALL = 'SendOnlyToAll'
-SEND_ONLY_TO_CHANGED = 'SendOnlyToChanged'
-SEND_TO_ALL_AND_SAVE_COPY = 'SendToAllAndSaveCopy'
-SEND_TO_CHANGED_AND_SAVE_COPY = 'SendToChangedAndSaveCopy'
-SEND_MEETING_INVITATIONS_CHOICES = (SEND_TO_NONE, SEND_ONLY_TO_ALL, SEND_TO_ALL_AND_SAVE_COPY)
-SEND_MEETING_INVITATIONS_AND_CANCELLATIONS_CHOICES = (SEND_TO_NONE, SEND_ONLY_TO_ALL, SEND_ONLY_TO_CHANGED,
-                                                      SEND_TO_ALL_AND_SAVE_COPY, SEND_TO_CHANGED_AND_SAVE_COPY)
-SEND_MEETING_CANCELLATIONS_CHOICES = (SEND_TO_NONE, SEND_ONLY_TO_ALL, SEND_TO_ALL_AND_SAVE_COPY)
-
-# AffectedTaskOccurrences values. See
-# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/deleteitem
-ALL_OCCURRENCIES = 'AllOccurrences'
-SPECIFIED_OCCURRENCE_ONLY = 'SpecifiedOccurrenceOnly'
-AFFECTED_TASK_OCCURRENCES_CHOICES = (ALL_OCCURRENCIES, SPECIFIED_OCCURRENCE_ONLY)
-
-# ConflictResolution values. See
-# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/updateitem
-NEVER_OVERWRITE = 'NeverOverwrite'
-AUTO_RESOLVE = 'AutoResolve'
-ALWAYS_OVERWRITE = 'AlwaysOverwrite'
-CONFLICT_RESOLUTION_CHOICES = (NEVER_OVERWRITE, AUTO_RESOLVE, ALWAYS_OVERWRITE)
-
-# DeleteType values. See
-# https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/deleteitem
-HARD_DELETE = 'HardDelete'
-SOFT_DELETE = 'SoftDelete'
-MOVE_TO_DELETED_ITEMS = 'MoveToDeletedItems'
-DELETE_TYPE_CHOICES = (HARD_DELETE, SOFT_DELETE, MOVE_TO_DELETED_ITEMS)
-
 
 class Item(BaseItem):
     """
@@ -53,7 +20,7 @@ class Item(BaseItem):
     """
     ELEMENT_NAME = 'Item'
 
-    LOCAL_FIELDS = [
+    LOCAL_FIELDS = Fields(
         MimeContentField('mime_content', field_uri='item:MimeContent', is_read_only_after_send=True),
         EWSElementField('parent_folder_id', field_uri='item:ParentFolderId', value_cls=ParentFolderId,
                         is_read_only=True),
@@ -86,8 +53,8 @@ class Item(BaseItem):
         BooleanField('reminder_is_set', field_uri='item:ReminderIsSet', is_required=True, default=False),
         IntegerField('reminder_minutes_before_start', field_uri='item:ReminderMinutesBeforeStart',
                      is_required_after_save=True, min=0, default=0),
-        CharField('display_cc', field_uri='item:DisplayCc', is_read_only=True),
-        CharField('display_to', field_uri='item:DisplayTo', is_read_only=True),
+        TextField('display_cc', field_uri='item:DisplayCc', is_read_only=True),
+        TextField('display_to', field_uri='item:DisplayTo', is_read_only=True),
         BooleanField('has_attachments', field_uri='item:HasAttachments', is_read_only=True),
         # ExtendedProperty fields go here
         CultureField('culture', field_uri='item:Culture', is_required_after_save=True, is_searchable=False),
@@ -102,8 +69,7 @@ class Item(BaseItem):
         EWSElementField('conversation_id', field_uri='item:ConversationId', value_cls=ConversationId,
                         is_read_only=True, supported_from=EXCHANGE_2010),
         BodyField('unique_body', field_uri='item:UniqueBody', is_read_only=True, supported_from=EXCHANGE_2010),
-    ]
-
+    )
     FIELDS = LOCAL_FIELDS[0:1] + BaseItem.FIELDS + LOCAL_FIELDS[1:]
 
     __slots__ = tuple(f.name for f in LOCAL_FIELDS)
@@ -133,20 +99,22 @@ class Item(BaseItem):
                 conflict_resolution=conflict_resolution,
                 send_meeting_invitations=send_meeting_invitations
             )
-            if self.id != item_id:
+            if self.id != item_id and not isinstance(self._id, (OccurrenceItemId, RecurringMasterItemId)):
+                # When we update an item with an OccurrenceItemId as ID, EWS returns the ID of the occurrence, so
+                # the ID of this item changes.
                 raise ValueError("'id' mismatch in returned update response")
             # Don't check that changekeys are different. No-op saves will sometimes leave the changekey intact
-            self.changekey = changekey
+            self._id = self.ID_ELEMENT_CLS(item_id, changekey)
         else:
             if update_fields:
                 raise ValueError("'update_fields' is only valid for updates")
             tmp_attachments = None
-            if self.account and self.account.version.build < EXCHANGE_2010 and self.attachments:
-                # Exchange 2007 can't save attachments immediately. You need to first save, then attach. Store
-                # the attachment of this item temporarily and attach later.
+            if self.account and self.account.version.build < EXCHANGE_2013 and self.attachments:
+                # At least some versions prior to Exchange 2013 can't save attachments immediately. You need to first
+                # save, then attach. Store the attachment of this item temporarily and attach later.
                 tmp_attachments, self.attachments = self.attachments, []
             item = self._create(message_disposition=SAVE_ONLY, send_meeting_invitations=send_meeting_invitations)
-            self.id, self.changekey = item.id, item.changekey
+            self._id = self.ID_ELEMENT_CLS(item.id, item.changekey)
             for old_att, new_att in zip(self.attachments, item.attachments):
                 if old_att.attachment_id is not None:
                     raise ValueError("Old 'attachment_id' is not empty")
@@ -158,22 +126,19 @@ class Item(BaseItem):
                 self.attach(tmp_attachments)
         return self
 
+    @require_account
     def _create(self, message_disposition, send_meeting_invitations):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        # bulk_create() returns an Item because we want to return the ID of both the main item *and* attachments
-        res = self.account.bulk_create(
-            items=[self], folder=self.folder, message_disposition=message_disposition,
-            send_meeting_invitations=send_meeting_invitations)
-        if message_disposition in (SEND_ONLY, SEND_AND_SAVE_COPY):
-            if res:
-                raise ValueError('Got a response in non-save mode')
-            return None
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        return res[0]
+        # Return a BulkCreateResult because we want to return the ID of both the main item *and* attachments
+        res = CreateItem(account=self.account).get(
+            items=[self],
+            folder=self.folder,
+            message_disposition=message_disposition,
+            send_meeting_invitations=send_meeting_invitations,
+            expect_result=message_disposition not in (SEND_ONLY, SEND_AND_SAVE_COPY),
+        )
+        if res is None:
+            return
+        return BulkCreateResult.from_xml(elem=res, account=self)
 
     def _update_fieldnames(self):
         from .contact import Contact, DistributionList
@@ -202,80 +167,70 @@ class Item(BaseItem):
             update_fieldnames.append(f.name)
         return update_fieldnames
 
+    @require_account
     def _update(self, update_fieldnames, message_disposition, conflict_resolution, send_meeting_invitations):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
         if not self.changekey:
             raise ValueError('%s must have changekey' % self.__class__.__name__)
         if not update_fieldnames:
             # The fields to update was not specified explicitly. Update all fields where update is possible
             update_fieldnames = self._update_fieldnames()
-        # bulk_update() returns a tuple
-        res = self.account.bulk_update(
-            items=[(self, update_fieldnames)], message_disposition=message_disposition,
+        res = UpdateItem(account=self.account).get(
+            items=[(self, update_fieldnames)],
+            message_disposition=message_disposition,
             conflict_resolution=conflict_resolution,
-            send_meeting_invitations_or_cancellations=send_meeting_invitations)
-        if message_disposition == SEND_AND_SAVE_COPY:
-            if res:
-                raise ValueError('Got a response in non-save mode')
-            return None
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        return res[0]
+            send_meeting_invitations_or_cancellations=send_meeting_invitations,
+            suppress_read_receipts=True,
+            expect_result=message_disposition != SEND_AND_SAVE_COPY,
+        )
+        if res is None:
+            return
+        return Item.id_from_xml(res)
 
+    @require_id
     def refresh(self):
         # Updates the item based on fresh data from EWS
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
-        res = list(self.account.fetch(ids=[self]))
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        fresh_item = res[0]
-        if self.id != fresh_item.id:
-            raise ValueError('Unexpected ID of fresh item')
+        from ..folders import Folder
+        additional_fields = {
+            FieldPath(field=f) for f in Folder(root=self.account.root).allowed_item_fields(version=self.account.version)
+        }
+
+        elem = GetItem(account=self.account).get(items=[self], additional_fields=additional_fields, shape=ID_ONLY)
+        res = Folder.item_model_from_tag(elem.tag).from_xml(elem=elem, account=self.account)
+        if self.id != res.id and not isinstance(self._id, (OccurrenceItemId, RecurringMasterItemId)):
+            # When we refresh an item with an OccurrenceItemId as ID, EWS returns the ID of the occurrence, so
+            # the ID of this item changes.
+            raise ValueError("'id' mismatch in returned update response")
         for f in self.FIELDS:
-            setattr(self, f.name, getattr(fresh_item, f.name))
+            setattr(self, f.name, getattr(res, f.name))
         # 'parent_item' should point to 'self', not 'fresh_item'. That way, 'fresh_item' can be garbage collected.
         for a in self.attachments:
             a.parent_item = self
-        del fresh_item
+        del res
 
+    @require_id
     def copy(self, to_folder):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
-        res = self.account.bulk_copy(ids=[self], to_folder=to_folder)
-        if not res:
+        res = CopyItem(account=self.account).get(
+            items=[self],
+            to_folder=to_folder,
+            expect_result=None,
+        )
+        if res is None:
             # Assume 'to_folder' is a public folder or a folder in a different mailbox
             return
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        return res[0]
+        return Item.id_from_xml(res)
 
+    @require_id
     def move(self, to_folder):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
-        res = self.account.bulk_move(ids=[self], to_folder=to_folder)
-        if not res:
+        res = MoveItem(account=self.account).get(
+            items=[self],
+            to_folder=to_folder,
+            expect_result=None,
+        )
+        if res is None:
             # Assume 'to_folder' is a public folder or a folder in a different mailbox
-            self.id, self.changekey = None, None
+            self._id = None
             return
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        self.id, self.changekey = res[0]
+        self._id = self.ID_ELEMENT_CLS(*Item.id_from_xml(res))
         self.folder = to_folder
 
     def move_to_trash(self, send_meeting_cancellations=SEND_TO_NONE, affected_task_occurrences=ALL_OCCURRENCIES,
@@ -283,7 +238,7 @@ class Item(BaseItem):
         # Delete and move to the trash folder.
         self._delete(delete_type=MOVE_TO_DELETED_ITEMS, send_meeting_cancellations=send_meeting_cancellations,
                      affected_task_occurrences=affected_task_occurrences, suppress_read_receipts=suppress_read_receipts)
-        self.id, self.changekey = None, None
+        self._id = None
         self.folder = self.account.trash
 
     def soft_delete(self, send_meeting_cancellations=SEND_TO_NONE, affected_task_occurrences=ALL_OCCURRENCIES,
@@ -291,7 +246,7 @@ class Item(BaseItem):
         # Delete and move to the dumpster, if it is enabled.
         self._delete(delete_type=SOFT_DELETE, send_meeting_cancellations=send_meeting_cancellations,
                      affected_task_occurrences=affected_task_occurrences, suppress_read_receipts=suppress_read_receipts)
-        self.id, self.changekey = None, None
+        self._id = None
         self.folder = self.account.recoverable_items_deletions
 
     def delete(self, send_meeting_cancellations=SEND_TO_NONE, affected_task_occurrences=ALL_OCCURRENCIES,
@@ -299,32 +254,21 @@ class Item(BaseItem):
         # Remove the item permanently. No copies are stored anywhere.
         self._delete(delete_type=HARD_DELETE, send_meeting_cancellations=send_meeting_cancellations,
                      affected_task_occurrences=affected_task_occurrences, suppress_read_receipts=suppress_read_receipts)
-        self.id, self.changekey, self.folder = None, None, None
+        self._id, self.folder = None, None
 
+    @require_id
     def _delete(self, delete_type, send_meeting_cancellations, affected_task_occurrences, suppress_read_receipts):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
-        res = self.account.bulk_delete(
-            ids=[self], delete_type=delete_type, send_meeting_cancellations=send_meeting_cancellations,
-            affected_task_occurrences=affected_task_occurrences, suppress_read_receipts=suppress_read_receipts)
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
+        DeleteItem(account=self.account).get(
+            items=[self],
+            delete_type=delete_type,
+            send_meeting_cancellations=send_meeting_cancellations,
+            affected_task_occurrences=affected_task_occurrences,
+            suppress_read_receipts=suppress_read_receipts,
+        )
 
+    @require_id
     def archive(self, to_folder):
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
-        res = self.account.bulk_archive(ids=[self], to_folder=to_folder)
-        if len(res) != 1:
-            raise ValueError('Expected result length 1, but got %s' % res)
-        if isinstance(res[0], Exception):
-            raise res[0]
-        return res[0]
+        return ArchiveItem(account=self.account).get(items=[self], to_folder=to_folder)
 
     def attach(self, attachments):
         """Add an attachment, or a list of attachments, to this item. If the item has already been saved, the
@@ -365,12 +309,9 @@ class Item(BaseItem):
             if a in self.attachments:
                 self.attachments.remove(a)
 
+    @require_id
     def create_forward(self, subject, body, to_recipients, cc_recipients=None, bcc_recipients=None):
         from .message import ForwardItem
-        if not self.account:
-            raise ValueError('%s must have an account' % self.__class__.__name__)
-        if not self.id:
-            raise ValueError('%s must have an ID' % self.__class__.__name__)
         return ForwardItem(
             account=self.account,
             reference_item_id=ReferenceItemId(id=self.id, changekey=self.changekey),
@@ -389,19 +330,3 @@ class Item(BaseItem):
             cc_recipients,
             bcc_recipients,
         ).send()
-
-
-class BulkCreateResult(BaseItem):
-    """A dummy class to store return values from a CreateItem service call"""
-    LOCAL_FIELDS = [
-        AttachmentField('attachments', field_uri='item:Attachments'),  # ItemAttachment or FileAttachment
-    ]
-    FIELDS = BaseItem.FIELDS + LOCAL_FIELDS
-
-    __slots__ = tuple(f.name for f in LOCAL_FIELDS)
-
-    def __init__(self, **kwargs):
-        super().__init__(**kwargs)
-        # pylint: disable=access-member-before-definition
-        if self.attachments is None:
-            self.attachments = []
