@@ -132,7 +132,7 @@ class QuerySet(SearchableMixIn):
         raise InvalidField("Unknown field path %r on folders %s" % (field_path, self.folder_collection.folders))
 
     @property
-    def _item_id_field(self):
+    def _id_field(self):
         return self._get_field_path('id')
 
     @property
@@ -288,11 +288,19 @@ class QuerySet(SearchableMixIn):
             yield val
         self._cache = _cache
 
+    """Do not implement __len__. The implementation of list() tries to preallocate memory by calling __len__ on the
+    given sequence, before calling __iter__. If we implemented __len__, we would end up calling FindItems twice, once
+    to get the result of self.count(), an once to return the actual result.
+
+    Also, according to https://stackoverflow.com/questions/37189968/how-to-have-list-consume-iter-without-calling-len,
+    a __len__ implementation should be cheap. That does not hold for self.count().
+
     def __len__(self):
         if self.is_cached:
             return len(self._cache)
         # This queryset has no cache yet. Call the optimized counting implementation
         return self.count()
+    """
 
     def __getitem__(self, idx_or_slice):
         # Support indexing and slicing. This is non-greedy when possible (slicing start, stop and step are not negative,
@@ -350,7 +358,7 @@ class QuerySet(SearchableMixIn):
             # _query() will return an iterator of (id, changekey) tuples
             if self._changekey_field not in self.only_fields:
                 transform_func = id_only_func
-            elif self._item_id_field not in self.only_fields:
+            elif self._id_field not in self.only_fields:
                 transform_func = changekey_only_func
             else:
                 transform_func = id_and_changekey_func
@@ -544,7 +552,7 @@ class QuerySet(SearchableMixIn):
             # We allow calling get(id=..., changekey=...) to get a single item, but only if exactly these two
             # kwargs are present.
             account = self.folder_collection.account
-            item_id = self._item_id_field.field.clean(kwargs['id'], version=account.version)
+            item_id = self._id_field.field.clean(kwargs['id'], version=account.version)
             changekey = self._changekey_field.field.clean(kwargs.get('changekey'), version=account.version)
             items = list(account.fetch(ids=[(item_id, changekey)], only_fields=self.only_fields))
         else:
@@ -667,21 +675,22 @@ class QuerySet(SearchableMixIn):
     def __str__(self):
         fmt_args = [('q', str(self.q)), ('folders', '[%s]' % ', '.join(str(f) for f in self.folder_collection.folders))]
         if self.is_cached:
-            fmt_args.append(('len', str(len(self))))
+            fmt_args.append(('len', str(len(self._cache))))
         return self.__class__.__name__ + '(%s)' % ', '.join('%s=%s' % (k, v) for k, v in fmt_args)
 
 
 def _get_value_or_default(item, field_order):
     # Python can only sort values when <, > and = are implemented for the two types. Try as best we can to sort
     # items, even when the item may have a None value for the field in question, or when the item is an
-    # Exception. If the field to be sorted by does not have a default value, there's really nothing we can do
-    # about it; we'll eventually raise a TypeError. If it does, we sort all None values and exceptions as the
-    # default value.
+    # Exception. If the field to be sorted by does not have a default value, try creating an empty instance if the
+    # field value class. If that doesn't work, there's really nothing we can do about it; we'll raise an error. If
+    # it does, we sort all None values and exceptions as the default value.
+    default = field_order.field_path.field.default or field_order.field_path.field.value_cls()
     if isinstance(item, Exception):
-        return field_order.field_path.field.default
+        return default
     val = field_order.field_path.get_value(item)
     if val is None:
-        return field_order.field_path.field.default
+        return default
     return val
 
 
