@@ -15,6 +15,7 @@ from ..errors import EWSWarning, TransportError, SOAPError, ErrorTimeoutExpired,
     ErrorCannotDeleteTaskOccurrence, ErrorMimeContentConversionFailed, ErrorRecurrenceHasNoOccurrence, \
     ErrorNoPublicFolderReplicaAvailable, MalformedResponseError, ErrorExceededConnectionCount, \
     SessionPoolMinSizeReached, ErrorIncorrectSchemaVersion, ErrorInvalidRequest
+from ..properties import FieldURI, IndexedFieldURI, ExtendedFieldURI, ExceptionFieldURI
 from ..transport import wrap, extra_headers
 from ..util import chunkify, create_element, add_xml_child, get_xml_attr, to_xml, post_ratelimited, \
     xml_to_str, set_xml_value, SOAPNS, TNS, MNS, ENS, ParseError
@@ -59,6 +60,23 @@ class EWSService(metaclass=abc.ABCMeta):
     # def get_payload(self, **kwargs):
     #     raise NotImplementedError()
 
+    def get(self, expect_result=True, **kwargs):
+        # Calls the service but expects exactly one result, or None when expect_result=False, or None or exactly one
+        # result when expect_result=None.
+        res = list(self.call(**kwargs))
+        if expect_result is None and not res:
+            # Allow empty result
+            return
+        if expect_result is False:
+            if res:
+                raise ValueError('Expected result length 0, but got %r', res)
+            return
+        if len(res) != 1:
+            raise ValueError('Expected result length 1, but got %r' % res)
+        if isinstance(res[0], Exception):
+            raise res[0]
+        return res[0]
+
     def _get_elements(self, payload):
         while True:
             try:
@@ -110,35 +128,50 @@ class EWSService(metaclass=abc.ABCMeta):
         # guessing tango, but then the server may decide that any arbitrary legacy backend server may actually process
         # the request for an account. Prepare to handle ErrorInvalidSchemaVersionForMailboxVersion errors and set the
         # server version per-account.
+        from ..credentials import IMPERSONATION, OAuth2Credentials
         from ..version import API_VERSIONS
+        account_to_impersonate = None
+        timezone = None
+        primary_smtp_address = None
         if isinstance(self, EWSAccountService):
-            account = self.account
             version_hint = self.account.version
+            if self.account.access_type == IMPERSONATION:
+                account_to_impersonate = self.account.identity
+            timezone = self.account.default_timezone
+            primary_smtp_address = self.account.primary_smtp_address
         else:
-            account = None
             # We may be here due to version guessing in Protocol.version, so we can't use the Protocol.version property
             version_hint = self.protocol.config.version
+            if isinstance(self.protocol.credentials, OAuth2Credentials):
+                account_to_impersonate = self.protocol.credentials.identity
         api_versions = [version_hint.api_version] + [v for v in API_VERSIONS if v != version_hint.api_version]
         for api_version in api_versions:
-            log.debug('Trying API version %s for account %s', api_version, account)
+            log.debug('Trying API version %s', api_version)
             r, session = post_ratelimited(
                 protocol=self.protocol,
                 session=self.protocol.get_session(),
                 url=self.protocol.service_endpoint,
-                headers=extra_headers(account=account),
-                data=wrap(content=payload, api_version=api_version, account=account),
+                headers=extra_headers(primary_smtp_address=primary_smtp_address),
+                data=wrap(
+                    content=payload,
+                    api_version=api_version,
+                    account_to_impersonate=account_to_impersonate,
+                    timezone=timezone,
+                ),
                 allow_redirects=False,
                 stream=self.streaming,
             )
+            # TODO: We should only release the session when we have fully consumed the response, but that requires fully
+            # consuming the generator returned by _get_soap_messages. The caller may not always do that. This
+            # seems to work anyway.
+            self.protocol.release_session(session)
             if self.streaming:
                 # Let 'requests' decode raw data automatically
                 r.raw.decode_content = True
-            else:
-                # If we're streaming, we want to wait to release the session until we have consumed the stream.
-                self.protocol.release_session(session)
             try:
                 header, body = self._get_soap_parts(response=r, **parse_opts)
             except ParseError as e:
+                r.close()  # Release memory
                 raise SOAPError('Bad SOAP response: %s' % e)
             # The body may contain error messages from Exchange, but we still want to collect version info
             if header is not None:
@@ -148,31 +181,24 @@ class EWSService(metaclass=abc.ABCMeta):
                 except TransportError as te:
                     log.debug('Failed to update version info (%s)', te)
             try:
-                res = self._get_soap_messages(body=body, **parse_opts)
+                return self._get_soap_messages(body=body, **parse_opts)
             except (ErrorInvalidServerVersion, ErrorIncorrectSchemaVersion, ErrorInvalidRequest):
                 # The guessed server version is wrong. Try the next version
                 log.debug('API version %s was invalid', api_version)
                 continue
             except ErrorInvalidSchemaVersionForMailboxVersion:
-                if not account:
-                    # This should never happen for non-account services
-                    raise ValueError("'account' should not be None")
                 # The guessed server version is wrong for this account. Try the next version
-                log.debug('API version %s was invalid for account %s', api_version, account)
+                log.debug('API version %s was invalid', api_version)
                 continue
             except ErrorExceededConnectionCount as e:
                 # ErrorExceededConnectionCount indicates that the connecting user has too many open TCP connections to
                 # the server. Decrease our session pool size.
-                if self.streaming:
-                    # In streaming mode, we haven't released the session yet, so we can't discard the session
-                    raise
-                else:
-                    try:
-                        self.protocol.decrease_poolsize()
-                        continue
-                    except SessionPoolMinSizeReached:
-                        # We're already as low as we can go. Let the user handle this.
-                        raise e
+                try:
+                    self.protocol.decrease_poolsize()
+                    continue
+                except SessionPoolMinSizeReached:
+                    # We're already as low as we can go. Let the user handle this.
+                    raise e
             except (ErrorTooManyObjectsOpened, ErrorTimeoutExpired) as e:
                 # ErrorTooManyObjectsOpened means there are too many connections to the Exchange database. This is very
                 # often a symptom of sending too many requests.
@@ -188,14 +214,12 @@ class EWSService(metaclass=abc.ABCMeta):
                 # Re-raise as an ErrorServerBusy with a default delay of 5 minutes
                 raise ErrorServerBusy(msg='Reraised from %s(%s)' % (e.__class__.__name__, e), back_off=300)
             finally:
-                if self.streaming:
-                    # TODO: We shouldn't release the session yet if we still haven't fully consumed the stream. It seems
-                    # a Session can handle multiple unfinished streaming requests, though.
-                    self.protocol.release_session(session)
-            return res
-        if account:
-            raise ErrorInvalidSchemaVersionForMailboxVersion('Tried versions %s but all were invalid for account %s' %
-                                                             (api_versions, account))
+                if not self.streaming:
+                    # In streaming mode, we may not have accessed the raw stream yet. Caller must handle this.
+                    r.close()  # Release memory
+
+        if isinstance(self, EWSAccountService):
+            raise ErrorInvalidSchemaVersionForMailboxVersion('Tried versions %s but all were invalid' % api_versions)
         raise ErrorInvalidServerVersion('Tried versions %s but all were invalid' % api_versions)
 
     def _handle_backoff(self, e):
@@ -339,17 +363,18 @@ class EWSService(metaclass=abc.ABCMeta):
         except self.ERRORS_TO_CATCH_IN_RESPONSE as e:
             return e
 
-    @classmethod
-    def _get_exception(cls, code, text, msg_xml):
+    @staticmethod
+    def _get_exception(code, text, msg_xml):
         if not code:
             return TransportError('Empty ResponseCode in ResponseMessage (MessageText: %s, MessageXml: %s)' % (
                 text, msg_xml))
         if msg_xml is not None:
             # If this is an ErrorInvalidPropertyRequest error, the xml may contain a specific FieldURI
-            for tag_name in ('FieldURI', 'IndexedFieldURI', 'ExtendedFieldURI', 'ExceptionFieldURI'):
-                field_uri_elem = msg_xml.find('{%s}%s' % (TNS, tag_name))
-                if field_uri_elem is not None:
-                    text += ' (field: %s)' % xml_to_str(field_uri_elem)
+            for elem_cls in (FieldURI, IndexedFieldURI, ExtendedFieldURI, ExceptionFieldURI):
+                elem = msg_xml.find(elem_cls.response_tag())
+                if elem is not None:
+                    field_uri = elem_cls.from_xml(elem, account=None)
+                    text += ' (field: %s)' % field_uri
             # If this is an ErrorInternalServerError error, the xml may contain a more specific error code
             inner_code, inner_text = None, None
             for value_elem in msg_xml.findall('{%s}Value' % TNS):
@@ -508,47 +533,24 @@ class PagingEWSMixIn(EWSService):
 class EWSPooledMixIn(EWSService):
     def _pool_requests(self, payload_func, items, **kwargs):
         log.debug('Processing items in chunks of %s', self.chunk_size)
-        # Chop items list into suitable pieces and let worker threads chew on the work. The order of the output result
-        # list must be the same as the input id list, so the caller knows which status message belongs to which ID.
-        # Yield results as they become available.
-        results = []
-        n = 0
-        for chunk in chunkify(items, self.chunk_size):
-            n += 1
-            log.debug('Starting %s._get_elements worker %s for %s items', self.__class__.__name__, n, len(chunk))
-            results.append((n, self.protocol.thread_pool.apply_async(
-                lambda c: self._get_elements(payload=payload_func(c, **kwargs)),
-                (chunk,)
-            )))
-
-            # Results will be available before iteration has finished if 'items' is a slow generator. Return early
-            while True:
-                if not results:
-                    break
-                i, r = results[0]
-                if not r.ready():
-                    # First non-yielded result isn't ready yet. Yielding other ready results would mess up ordering
-                    break
-                log.debug('%s._get_elements result %s is ready early', self.__class__.__name__, i)
-                for elem in r.get():
-                    yield elem
-                # Results object has been processed. Remove from list.
-                del results[0]
-
-        # Yield remaining results in order, as they become available
-        for i, r in results:
-            log.debug('Waiting for %s._get_elements result %s of %s', self.__class__.__name__, i, n)
-            elems = r.get()
-            log.debug('%s._get_elements result %s of %s is ready', self.__class__.__name__, i, n)
-            for elem in elems:
+        # Chop items list into suitable pieces. The order of the output result list must be the same as the input id
+        # list, so the caller knows which status message belongs to which ID. Yield results as they become available.
+        for i, chunk in enumerate(chunkify(items, self.chunk_size), start=1):
+            log.debug('Processing %s chunk %s containing %s items', self.__class__.__name__, i, len(chunk))
+            for elem in self._get_elements(payload=payload_func(chunk, **kwargs)):
                 yield elem
 
 
-def to_item_id(item, item_cls):
+def to_item_id(item, item_cls, version):
     # Coerce a tuple, dict or object to an 'item_cls' instance. Used to create [Parent][Item|Folder]Id instances from a
     # variety of input.
     if isinstance(item, item_cls):
+        # Allow any subclass of item_cls, e.g. OccurrenceItemId when ItemId is passed
         return item
+    from ..folders import BaseFolder
+    from ..items import BaseItem
+    if isinstance(item, (BaseFolder, BaseItem)):
+        return item.to_id_xml(version=version)
     if isinstance(item, (tuple, list)):
         return item_cls(*item)
     if isinstance(item, dict):
@@ -568,12 +570,12 @@ def create_shape_element(tag, shape, additional_fields, version):
 
 
 def create_folder_ids_element(tag, folders, version):
-    from ..folders import BaseFolder, FolderId, DistinguishedFolderId
+    from ..folders import FolderId, DistinguishedFolderId
     folder_ids = create_element(tag)
     for folder in folders:
         log.debug('Collecting folder %s', folder)
-        if not isinstance(folder, (BaseFolder, FolderId, DistinguishedFolderId)):
-            folder = to_item_id(folder, FolderId)
+        if not isinstance(folder, DistinguishedFolderId):
+            folder = to_item_id(folder, FolderId, version=version)
         set_xml_value(folder_ids, folder, version=version)
     if not len(folder_ids):
         raise ValueError('"folders" must not be empty')
@@ -585,7 +587,7 @@ def create_item_ids_element(items, version):
     item_ids = create_element('m:ItemIds')
     for item in items:
         log.debug('Collecting item %s', item)
-        set_xml_value(item_ids, to_item_id(item, ItemId), version=version)
+        set_xml_value(item_ids, to_item_id(item, ItemId, version=version), version=version)
     if not len(item_ids):
         raise ValueError('"items" must not be empty')
     return item_ids
