@@ -3,6 +3,7 @@ from codecs import BOM_UTF8
 from collections import OrderedDict
 import datetime
 from decimal import Decimal
+from functools import wraps
 import io
 import itertools
 import logging
@@ -11,10 +12,9 @@ import socket
 from threading import get_ident
 import time
 from urllib.parse import urlparse
-import xml.sax.handler
+import xml.sax.handler  # nosec
 
-# Import _etree via defusedxml instead of directly from lxml.etree, to silence overly strict linters
-from defusedxml.lxml import parse, tostring, GlobalParserTLS, RestrictedElement, _etree
+import lxml.etree  # nosec
 from defusedxml.expatreader import DefusedExpatParser
 from defusedxml.sax import _InputSource
 import dns.resolver
@@ -31,7 +31,27 @@ from .errors import TransportError, RateLimitError, RedirectError, RelativeRedir
 log = logging.getLogger(__name__)
 
 
-class ParseError(_etree.ParseError):
+def require_account(f):
+    @wraps(f)
+    def wrapper(self, *args, **kwargs):
+        if not self.account:
+            raise ValueError('%s must have an account' % self.__class__.__name__)
+        return f(self, *args, **kwargs)
+    return wrapper
+
+
+def require_id(f):
+    @wraps(f)
+    def wrapper(self, *args, **kwargs):
+        if not self.account:
+            raise ValueError('%s must have an account' % self.__class__.__name__)
+        if not self.id:
+            raise ValueError('%s must have an ID' % self.__class__.__name__)
+        return f(self, *args, **kwargs)
+    return wrapper
+
+
+class ParseError(lxml.etree.ParseError):
     """Used to wrap lxml ParseError in our own class"""
     pass
 
@@ -60,7 +80,7 @@ ns_translation = OrderedDict([
     ('t', TNS),
 ])
 for item in ns_translation.items():
-    _etree.register_namespace(*item)
+    lxml.etree.register_namespace(*item)
 
 
 def is_iterable(value, generators_allowed=False):
@@ -124,8 +144,8 @@ def xml_to_str(tree, encoding=None, xml_declaration=False):
     if xml_declaration and not encoding:
         raise ValueError("'xml_declaration' is not supported when 'encoding' is None")
     if encoding:
-        return tostring(tree, encoding=encoding, xml_declaration=True)
-    return tostring(tree, encoding=str, xml_declaration=False)
+        return lxml.etree.tostring(tree, encoding=encoding, xml_declaration=True)
+    return lxml.etree.tostring(tree, encoding=str, xml_declaration=False)
 
 
 def get_xml_attr(tree, name):
@@ -191,7 +211,7 @@ def set_xml_value(elem, value, version):
     from .version import Version
     if isinstance(value, (str, bool, bytes, int, Decimal, datetime.time, EWSDate, EWSDateTime)):
         elem.text = value_to_xml_text(value)
-    elif isinstance(value, RestrictedElement):
+    elif isinstance(value, _element_class):
         elem.append(value)
     elif is_iterable(value, generators_allowed=True):
         for v in value:
@@ -201,7 +221,7 @@ def set_xml_value(elem, value, version):
                 if not isinstance(version, Version):
                     raise ValueError("'version' %r must be a Version instance" % version)
                 elem.append(v.to_xml(version=version))
-            elif isinstance(v, RestrictedElement):
+            elif isinstance(v, _element_class):
                 elem.append(v)
             elif isinstance(v, str):
                 add_xml_child(elem, 't:String', v)
@@ -228,12 +248,11 @@ def create_element(name, attrs=None, nsmap=None):
     if ':' in name:
         ns, name = name.split(':')
         name = '{%s}%s' % (ns_translation[ns], name)
-    elem = RestrictedElement(nsmap=nsmap)
+    elem = _forgiving_parser.makeelement(name, nsmap=nsmap)
     if attrs:
         # Try hard to keep attribute order, to ensure deterministic output. This simplifies testing.
         for k, v in attrs.items():
             elem.set(k, v)
-    elem.tag = name
     return elem
 
 
@@ -300,8 +319,8 @@ class StreamingBase64Parser(DefusedExpatParser):
         self.buffer = None
         self.element_found = None
 
-    def parse(self, source):
-        raw_source = source.raw
+    def parse(self, r):
+        raw_source = r.raw
         # Like upstream but yields the return value of self.feed()
         raw_source = prepare_input_source(raw_source)
         self.prepareParser(raw_source)
@@ -318,7 +337,7 @@ class StreamingBase64Parser(DefusedExpatParser):
             buffer = file.read(self._bufsize)
         # Any remaining data in self.buffer should be padding chars now
         self.buffer = None
-        source.close()
+        r.close()  # Release memory
         self.close()
         if not self.element_found:
             data = bytes(collected_data)
@@ -344,15 +363,12 @@ class StreamingBase64Parser(DefusedExpatParser):
         self.buffer = [remainder] if remainder else []
 
 
-class ForgivingParser(GlobalParserTLS):
-    parser_config = {
-        'resolve_entities': False,
-        'recover': True,  # This setting is non-default
-        'huge_tree': True,  # This setting enables parsing huge attachments, mime_content and other large data
-    }
-
-
-_forgiving_parser = ForgivingParser()
+_forgiving_parser = lxml.etree.XMLParser(
+    resolve_entities=False,  # This setting is recommended by lxml for safety
+    recover=True,  # This setting is non-default
+    huge_tree=True,  # This setting enables parsing huge attachments, mime_content and other large data
+)
+_element_class = _forgiving_parser.makeelement('x').__class__
 
 
 class BytesGeneratorIO(io.RawIOBase):
@@ -410,12 +426,11 @@ def to_xml(bytes_content):
         stream = io.BytesIO(bytes_content)
     else:
         stream = BytesGeneratorIO(bytes_content)
-    forgiving_parser = _forgiving_parser.getDefaultParser()
     try:
-        return parse(stream, parser=forgiving_parser)
+        res = lxml.etree.parse(stream, parser=_forgiving_parser)  # nosec
     except AssertionError as e:
         raise ParseError(e.args[0], '<not from file>', -1, 0)
-    except _etree.ParseError as e:
+    except lxml.etree.ParseError as e:
         if hasattr(e, 'position'):
             e.lineno, e.offset = e.position
         if not e.lineno:
@@ -428,7 +443,7 @@ def to_xml(bytes_content):
         else:
             offending_excerpt = offending_line[max(0, e.offset - 20):e.offset + 20]
             msg = '%s\nOffending text: [...]%s[...]' % (str(e), offending_excerpt)
-            raise ParseError(msg, e.lineno, e.offset)
+            raise ParseError(msg, '<not from file>', e.lineno, e.offset)
     except TypeError:
         try:
             stream.seek(0)
@@ -436,6 +451,15 @@ def to_xml(bytes_content):
             pass
         raise ParseError('This is not XML: %r' % stream.read(), '<not from file>', -1, 0)
 
+    if res.getroot() is None:
+        try:
+            stream.seek(0)
+            msg = 'No root element found: %r' % stream.read()
+        except (IndexError, io.UnsupportedOperation):
+            msg = 'No root element found'
+        raise ParseError(msg, '<not from file>', -1, 0)
+    return res
+
 
 def is_xml(text):
     """
@@ -452,12 +476,12 @@ class PrettyXmlHandler(logging.StreamHandler):
     """A steaming log handler that prettifies log statements containing XML when output is a terminal"""
     @staticmethod
     def parse_bytes(xml_bytes):
-        return parse(io.BytesIO(xml_bytes))
+        return lxml.etree.parse(io.BytesIO(xml_bytes), parser=_forgiving_parser)  # nosec
 
     @classmethod
     def prettify_xml(cls, xml_bytes):
         # Re-formats an XML document to a consistent style
-        return tostring(
+        return lxml.etree.tostring(
             cls.parse_bytes(xml_bytes),
             xml_declaration=True,
             encoding='utf-8',
@@ -505,7 +529,7 @@ class AnonymizingXmlHandler(PrettyXmlHandler):
         super().__init__(*args, **kwargs)
 
     def parse_bytes(self, xml_bytes):
-        root = parse(io.BytesIO(xml_bytes))
+        root = lxml.etree.parse(io.BytesIO(xml_bytes), parser=_forgiving_parser)  # nosec
         for elem in root.iter():
             for attr in set(elem.keys()) & {'RootItemId', 'ItemId', 'Id', 'RootItemChangeKey', 'ChangeKey'}:
                 elem.set(attr, 'DEADBEEF=')
@@ -520,17 +544,21 @@ class DummyRequest:
 
 
 class DummyResponse:
-    def __init__(self, url, headers, request_headers, content=b'', status_code=503):
+    def __init__(self, url, headers, request_headers, content=b'', status_code=503, history=None):
         self.status_code = status_code
         self.url = url
         self.headers = headers
         self.content = content
         self.text = content.decode('utf-8', errors='ignore')
         self.request = DummyRequest(headers=request_headers)
+        self.history = history
 
     def iter_content(self):
         return self.content
 
+    def close(self):
+        pass
+
 
 def get_domain(email):
     try:
@@ -708,10 +736,12 @@ Response data: %(xml_response)s
                 )
             log.debug(log_msg, log_vals)
             if _need_new_credentials(response=r):
+                r.close()  # Release memory
                 session = protocol.refresh_credentials(session)
                 continue
             total_wait = time.monotonic() - t_start
             if _may_retry_on_error(response=r, retry_policy=protocol.retry_policy, wait=total_wait):
+                r.close()  # Release memory
                 log.info("Session %s thread %s: Connection error on URL %s (code %s). Cool down %s secs",
                          session.session_id, thread_id, r.url, r.status_code, wait)
                 protocol.retry_policy.back_off(wait)
@@ -719,8 +749,7 @@ Response data: %(xml_response)s
                 wait *= 2  # Increase delay for every retry
                 continue
             if r.status_code in (301, 302):
-                if stream:
-                    r.close()
+                r.close()  # Release memory
                 url, redirects = _redirect_or_fail(r, redirects, allow_redirects)
                 continue
             break
@@ -738,11 +767,7 @@ Response data: %(xml_response)s
         log.debug('Got status code %s but trying to parse content anyway', r.status_code)
     elif r.status_code != 200:
         protocol.retire_session(session)
-        try:
-            _raise_response_errors(r, protocol, log_msg, log_vals)  # Always raises an exception
-        finally:
-            if stream:
-                r.close()
+        _raise_response_errors(r, protocol, log_msg, log_vals)  # Always raises an exception
     log.debug('Session %s thread %s: Useful response from %s', session.session_id, thread_id, url)
     return r, session
 
@@ -819,7 +844,7 @@ def _raise_response_errors(response, protocol, log_msg, log_vals):
         raise TransportError('The service account is currently locked out')
     if response.status_code == 401 and protocol.retry_policy.fail_fast:
         # This is a login failure
-        raise UnauthorizedError('Wrong username or password for %s' % response.url)
+        raise UnauthorizedError('Invalid credentials for %s' % response.url)
     if 'TimeoutException' in response.headers:
         raise response.headers['TimeoutException']
     # This could be anything. Let higher layers handle this. Add full context for better debugging.
