@@ -51,7 +51,6 @@ class ProtocolTest(EWSTest):
             self.assertEqual(base_p, p)
             self.assertEqual(id(base_p), id(p))
             self.assertEqual(hash(base_p), hash(p))
-            self.assertEqual(id(base_p.thread_pool), id(p.thread_pool))
             self.assertEqual(id(base_p._session_pool), id(p._session_pool))
 
     def test_close(self):
@@ -72,14 +71,18 @@ class ProtocolTest(EWSTest):
         self.assertEqual(len({p.raddr[0] for p in proc.connections() if p.raddr[0] in ip_addresses}), 0)
 
     def test_poolsize(self):
-        self.assertEqual(self.account.protocol.SESSION_POOLSIZE, 4)
+        self.assertEqual(self.account.protocol.SESSION_POOLSIZE, 1)
 
     def test_decrease_poolsize(self):
+        # Temporarily change the session pool size so we can test decreasing the pool size
+        tmp = Protocol.SESSION_POOLSIZE
+        Protocol.SESSION_POOLSIZE = 4
         protocol = Protocol(config=Configuration(
             service_endpoint='https://example.com/Foo.asmx', credentials=Credentials('A', 'B'),
             auth_type=NTLM, version=Version(Build(15, 1)), retry_policy=FailFast()
         ))
-        self.assertEqual(protocol._session_pool.qsize(), Protocol.SESSION_POOLSIZE)
+        Protocol.SESSION_POOLSIZE = tmp
+        self.assertEqual(protocol._session_pool.qsize(), 4)
         protocol.decrease_poolsize()
         self.assertEqual(protocol._session_pool.qsize(), 3)
         protocol.decrease_poolsize()
@@ -110,7 +113,7 @@ class ProtocolTest(EWSTest):
         accounts = [(self.account, 'Organizer', False)]
 
         with self.assertRaises(ValueError):
-            self.account.protocol.get_free_busy_info(accounts=[('XXX', 'XXX', 'XXX')], start=0, end=0)
+            self.account.protocol.get_free_busy_info(accounts=[(123, 'XXX', 'XXX')], start=0, end=0)
         with self.assertRaises(ValueError):
             self.account.protocol.get_free_busy_info(accounts=[(self.account, 'XXX', 'XXX')], start=0, end=0)
         with self.assertRaises(ValueError):
@@ -129,6 +132,12 @@ class ProtocolTest(EWSTest):
             ms_id = view_info.working_hours_timezone.to_server_timezone(server_timezones, start.year)
             self.assertIn(ms_id, {t[0] for t in CLDR_TO_MS_TIMEZONE_MAP.values()})
 
+        # Test account as simple email
+        for view_info in self.account.protocol.get_free_busy_info(
+                accounts=[(self.account.primary_smtp_address, 'Organizer', False)], start=start, end=end
+        ):
+            self.assertIsInstance(view_info, FreeBusyView)
+
     def test_get_roomlists(self):
         # The test server is not guaranteed to have any room lists which makes this test less useful
         ws = GetRoomLists(self.account.protocol)
@@ -432,7 +441,7 @@ class ProtocolTest(EWSTest):
         self.assertEqual(len(return_ids), len(items))
         ids = self.account.calendar.filter(start__lt=end, end__gt=start, categories__contains=self.categories) \
             .values_list('id', 'changekey')
-        self.assertEqual(len(ids), len(items))
+        self.assertEqual(ids.count(), len(items))
 
     def test_disable_ssl_verification(self):
         # Test that we can make requests when SSL verification is turned off. I don't know how to mock TLS responses
